# -*- coding: latin-1 -*-


from ftw.pdfgenerator.html2latex.converter import HTML2LatexConverter
from ftw.pdfgenerator.html2latex.subconverters import textformatting
from ftw.pdfgenerator.html2latex.subconverters import url
from unittest2 import TestCase


class BasicConverter(HTML2LatexConverter):
    """A HTML2LatexConverter where subconverters are disabled.
    """

    def get_default_subconverters(self):
        return [url.URLConverter,
                textformatting.Textformatting]


class TestBasicPatterns(TestCase):

    def setUp(self):
        super(TestBasicPatterns, self).setUp()

        self.converter = BasicConverter(
            context=object(),
            request=object(),
            layout=object())

        self.convert = self.converter.convert

    def test_common_text_formatting(self):
        self.assertEqual(self.convert('Hello <b>World</b>!'),
                         'Hello \\textbf{World}!')

        self.assertEqual(self.convert('Hello <strong>World</strong>!'),
                         'Hello \\textbf{World}!')

        self.assertEqual(self.convert('Hello <u>World</u>!'),
                         'Hello \\emph{World}!')

        self.assertEqual(self.convert('Hello <u id="foo">World</u>!'),
                         'Hello \\emph{World}!')

        self.assertEqual(self.convert('Hello <em>World</em>!'),
                         'Hello \\emph{World}!')

        self.assertEqual(self.convert('Hello <em id="foo">World</em>!'),
                         'Hello \\emph{World}!')

        self.assertEqual(self.convert('Hello <i>World</i>!'),
                         'Hello \\textit{World}!')

        self.assertEqual(self.convert('Hello <i id="foo">World</i>!'),
                         'Hello \\textit{World}!')

        self.assertEqual(self.convert('capacity: 55 m<sup>3</sup>'),
                         'capacity: 55 m\\textsuperscript{3}')

    def test_remove_unused_tags(self):
        self.assertEqual(self.convert('<span>Hello World!</span>'),
                         'Hello World!')

        self.assertEqual(self.convert('Hallo <b></b> Welt <asdf></asdf>'),
                         'Hallo  Welt')

        # Remove image maps
        self.assertEqual(self.convert(
                'foo\n'
                '<map name="ImageMap1">\n'

                '<area shape="rect" coords="158, 128, 276, 194"'
                ' href="http://localhost//bar-2005-2015-31.12.2009"'
                ' target="_self" alt="Bar">\n'

                '<area shape="rect" coords="170, 333, 287, 398"'
                ' href="http://localhost//foo" target="_self" alt="foo">\n'

                '<area shape="rect" coords="310, 137, 421, 179"'
                ' href="http://localhost//foo-bar" target="_self"'
                ' alt="Foo B������r">\n'

                '<area shape="rect" coords="36, 168, 123, 208" '
                'href="./a/url" target="_blank" alt="Something">\n'
                '</map>\n'
                'bar'),
                         'foo       bar')

    def test_headings(self):
        self.assertEqual(self.convert('<h1>My Heading</h1>'),
                         '\\section{My Heading}')

        self.assertEqual(self.convert('<h2 align="right">Subheading</h2>'),
                         '\\subsection{Subheading}')

    def test_quotation_marks(self):
        self.assertEqual(
            self.convert('Hello "World" that\'s &quot;my&quot; test..'),
            'Hello "`World"\' that\'s "`my"\' test..')

        self.assertEqual(
            self.convert('Hello "World" that\'s &quot;my&quot;, test..'),
            'Hello "`World"\' that\'s "`my"\', test..')

        self.assertEqual(self.convert('"Hello" World'),
                         '"`Hello"\' World')

        self.assertEqual(self.convert('Bar "foo (1)" blubb'),
                         'Bar "`foo (1)"\' blubb')

        self.assertEqual('"`Hello World"\'',
                         self.convert('<p>"Hello World"</p>'))

    def test_quotation_marks2(self):
        self.assertEqual(self.convert('foo &quot;<b>bar</b>&quot; baz'),
                         'foo "`\\textbf{bar}"\' baz')

    def test_quotation_marks3(self):
        self.assertEqual(self.convert('foo <b>&quot;bar&quot;</b> baz'),
                         'foo \\textbf{"`bar"\'} baz')

    def test_whitespace(self):
        self.assertEqual(self.convert('U\x0bV\x0cW\r\nX\nY\rZ'),
                         'U V W X Y Z')

        # Use short whitespace (\,) for abbreviation:
        self.assertEqual(self.convert('x e.g. y').strip(),
                         'x e.\,g. y')

        self.assertEqual(self.convert('x e. g. y').strip(),
                         'x e.\,g. y')

        self.assertEqual(self.convert('test i. d. r. foo').strip(),
                         'test i.\,d.\,r. foo')

        # test trimming newlines
        self.assertEqual(self.convert('<b>eins</b>'),
                         '\\textbf{eins}')

        self.assertEqual(self.convert('<b>zwei</b><br />zwei'),
                         '\\textbf{zwei}\\\\\nzwei')

        self.assertEqual(self.convert('<b>drei<br /></b>'),
                         '\\textbf{drei\\\\\n}')

        self.assertEqual(self.convert('<b>vier<br />\n</b>'),
                         '\\textbf{vier\\\\\n }')

        self.assertEqual(self.convert('<b>fuenf<br />\n</b><br />'),
                         '\\textbf{fuenf\\\\\n }')

        self.assertEqual(self.convert('<b>sechs<br />\n</b>sechs<br />'),
                         '\\textbf{sechs\\\\\n }sechs')

        self.assertEqual(self.convert('<strong>&quot;sieben&quot;<br />\n'
                                      '</strong><br />'),
                         '\\textbf{"`sieben"\'\\\\\n }')

    def test_abbreviations_are_not_breakable(self):
        self.maxDiff = None

        self.assert_convertions({
                'S. 5': 'S.~5',
                'S. 15': 'S.~15',
                'S. 17. Bar': 'S.~17. Bar',
                'Sa 2': 'Sa 2',

                'Abs. 5': 'Abs.~5',
                'Abs. 15': 'Abs.~15',
                'Abs. 17. Bar': 'Abs.~17. Bar',
                'Absx 2': 'Absx 2',

                '5 m2': '5~m2',
                '17 m2': '17~m2',
                'a5 m2': 'a5~m2',

                'Art. 5': 'Art.~5',
                'Art. 15': 'Art.~15',
                'Art. 17. Bar': 'Art.~17. Bar',
                'Arta 1': 'Arta 1',

                'SR 5': 'SR~5',
                'SR 15': 'SR~15',
                'SR 17. Bar': 'SR~17. Bar',
                })

    def test_paragraphs_and_newlines(self):
        self.assertEqual(self.convert('Hello<br><br>World'),
                         'Hello\n\nWorld')

        self.assertEqual(self.convert('Hello<br /> \n  <br/>World'),
                         'Hello\n\nWorld')

        self.assertEqual(self.convert('<p>Hello</p><p>World</p>'),
                         'Hello\n\nWorld')

        self.assertEqual(self.convert('Hello<br>World'),
                         'Hello\\\\\nWorld')

        self.assertEqual(self.convert('Carriage<br />Return'),
                         'Carriage\\\\\nReturn')

        self.assertEqual(
            self.convert('<p>Hello</p><p>World</p><p></p><p></p><p></p>'),
            'Hello\n\nWorld')

        self.assertEqual(
            self.convert('<p><span style="font-weight: bold;">Mitglieder'
                         '</span><br />Name1<br />Name2<br /></p>').strip(),
            '\n'.join((
                    r'\textbf{Mitglieder}\\',
                    r'Name1\\',
                    r'Name2')))

        self.assertEqual(self.convert('<p><strong><br />Text</strong></p>'),
                         r'\textbf{Text}')

    def test_curly_bracket_spaces(self):
        # If there is no space at left of the curly brackets, there will
        # not be any space in the displayed text (PDF), but in HTML it
        # is (with the bold-tag in HTML the browser displays a bold space).

        self.assertEqual(self.convert('text<b> bold text</b>'),
                         'text \\textbf{bold text}')

        self.assertEqual(self.convert('text<strong> bold text</strong>'),
                         'text \\textbf{bold text}')

        self.assertEqual(self.convert('te<b>X</b>t'),
                         'te\\textbf{X}t')

        self.assertEqual(self.convert('a<em> b</em> c'),
                         'a \\emph{b} c')

        self.assertEqual(self.convert('a<u> b</u> c'),
                         'a \\emph{b} c')

        self.assertEqual(self.convert('a<i> b</i> c'),
                         'a \\textit{b} c')

        self.assertEqual(self.convert('a <i>b</i> c'),
                         'a \\textit{b} c')

        self.assertEqual(self.convert('a<i>b</i>c'),
                         'a\\textit{b}c')

    def test_links(self):
        # Blank links should be stripped
        self.assertEqual(
            self.convert('<a href="about:blank">Something</a>'),
            'Something')

    def test_fckeditor_compatibility(self):
        # FCKEditor doesnt use <b> nor <strong> but <span>
        # with style-attribute:
        self.assertEqual(
            self.convert('Hello <span style="font-weight: bold;">my</span>'
                         ' world'),
            'Hello \\textbf{my} world')

        self.assertEqual(
            self.convert('<span style="font-weight: bold;">test</span>'),
            '\\textbf{test}')

        self.assertEqual(
            self.convert('<span style="font-weight: bold">test</span>'),
            '\\textbf{test}')

    def test_asterisks(self):
        """Asterisks (*) needs to be defined using the \ast command,
        which does only work properly in math-environments ($$).
        """
        self.assertEqual(self.convert('foo * bar'),
                         r'foo $\ast$ bar')

    def test_sepcial_characters(self):
        self.assertEqual(self.convert('!#$%&amp;\'()+-./02345'),
                         "!\\#\\$\\%\\&'()+-./02345")

        self.assertEqual(self.convert('6789:;&lt;=&gt;?@ABCDEFGHI'),
                         '6789:;<=>?@ABCDEFGHI')

        self.assertEqual(self.convert('JKLMNOPQRSTUVWXYZ[]^'),
                         'JKLMNOPQRSTUVWXYZ[]\\^{}')

        self.assertEqual(self.convert('_`abcdefghiklmnopqr'),
                         '\\_`abcdefghiklmnopqr')

        self.assertEqual(self.convert('stuvwxyz{|}~'),
                         'stuvwxyz\\{|\\}\\~{}')

        # XXX: fix those tests: maybe decode to unicode or use propper utf8
        # char escaping.
        self.assertEqual(
            self.convert('&euro;&lsquo;&rsquo;&rsquo;'
                         '&ldquo;&rdquo;&ndash;&mdash;'),
            '\euro{}‘’’�'
            '��”–—')

        self.assertEqual(
            self.convert(
                '&iexcl;&cent;&pound;&curren;&sect;&uml;&copy;&ordf;'),
            '¡\cent{}£\currency{}\S¨©ª')

        self.assertEqual(
            self.convert('&laquo;&reg;&macr;&sup2;&sup3;&acute;&micro;'),
            '«®¯²³´µ')

        self.assertEqual(
            self.convert('&middot;&cedil;&sup1;&ordm;&raquo;&iquest;'),
            '·¸¹º»¿')

        self.assertEqual(
            self.convert('ÀÁÂÃÅÆÇÈÉ'),
            'ÀÁÂÃÅÆÇÈÉ')

        self.assertEqual(
            self.convert('ÊËÌÍÎÏÐÑÒÓ'),
            'ÊËÌÍÎÏÐÑÒÓ')

        self.assertEqual(
            self.convert('ÔÕÖØÙÚÛÜÝ'),
            'ÔÕÖØÙÚÛÜÝ')

        self.assertEqual(
            self.convert('Þßàáâãäåæç'),
            'Þßàáâãäåæç')

        self.assertEqual(
            self.convert('èéêëìíîïðñ'),
            'èéêëìíîïðñ')

        self.assertEqual(
            self.convert('òóôõöøùúû'),
            'òóôõöøùúû')

        self.assertEqual(
            self.convert('üüýþÿŒœ&sbquo;&bdquo;'),
            'üüýþÿŒœ‚„')

        self.assertEqual(self.convert('&divide;'), '$\\div$')
        self.assertEqual(self.convert('&times;'), '$\\times$')
        self.assertEqual(self.convert('&frac12;'), '$\\frac12$')
        self.assertEqual(self.convert('&frac34;'), '$\\frac34$')
        self.assertEqual(self.convert('&frac14;'), '$\\frac14$')
        self.assertEqual(self.convert('&para;'), '\\P')
        self.assertEqual(self.convert('&deg;'), '$\\circ$')
        self.assertEqual(self.convert('&plusmn;'), '$\\pm$')
        self.assertEqual(self.convert('&not;'), '\\ensuremath{\\lnot}')
        self.assertEqual(self.convert('&hellip;'), '$\\dots$')
        self.assertEqual(self.convert('&trade;'), '™')
        self.assertEqual(self.convert('&bull;'), '•')
        self.assertEqual(self.convert('&rarr;'), '$\\rightarrow$')
        self.assertEqual(self.convert('&rArr;'), '$\\Rightarrow$')
        self.assertEqual(self.convert('&hArr;'), '⇔')
        self.assertEqual(self.convert('&asymp;'), '$\\approx$')

        # unsupported sepcial chars:
        self.assertEqual(
            self.convert('&yen;&brvbar;‛&diams;►'),
            '')

    def test_hyphenation_characters(self):
        self.assertEqual(
            self.convert('ergo\xc2\xadnomic'), 'ergo"-nomic')
        self.assertEqual(
            self.convert('ergo&shy;nomic'), 'ergo"-nomic')

    def test_non_breakable_hyphens_before_comma(self):
        self.assert_convertions({
                'foo-, bar': 'foo"~, bar',
                'foo- bar': 'foo"= bar'})

    def test_remove_css(self):
        self.assertEqual(self.convert(
                'hello\n'
                '<style type="text/css">\n'
                '    <!--\n'
                '        body {\n'
                '            background-color: red;\n'
                '        }\n'
                '    /-->\n'
                '</style>\n'
                'world').strip(),
                         'hello  world')

    def test_various(self):
        # Delimiter Symbol must should be escaped in LaTeX-Style:
        self.assertEqual(
            self.convert('Unterstützungs- und Hilfsberichte').strip(),
            'Unterstützungs"= und Hilfsberichte')

        # We don't like linebreaks between numbers and a %-character,
        # so we prevent it
        self.assertEqual(self.convert('101.5 %').strip(),
                         '101.5\,\%')

        # But maybe we use a %-character in the text:
        self.assertEqual(self.convert('we use %-characters').strip(),
                         'we use \%-characters')

        # Test some german stuff
        self.assertEqual(
            self.convert('foo lic.iur. foo Dr.iur. foo z.B. foo '
                         'z. B. foo').strip(),
            'foo lic.\,iur. foo Dr.\,iur. foo z.\,B. foo z.\,B. foo')

        # We dont want spaces to break:
        self.assertEqual(self.convert('1. Januar 2008').strip(),
                         '1.~Januar 2008')

        self.assertEqual(self.convert('4. Feb. 2008').strip(),
                         '4.~Feb. 2008')

        # Usually we want to have a small space after a paragraph character,
        # if it is followed by a integer (\S is the LaTeX command for
        # the §-character):
        self.assertEqual(self.convert('foo §1 foo § 2 foo').strip(),
                         'foo \S\,1 foo \S\,2 foo')

        # Backslashes
        self.assertEqual(self.convert(r'C:\Programs\foo').strip(),
                         r'C:\textbackslash Programs\textbackslash foo')

    def test_removes_nonbreaking_spaces(self):
        # Non break spaces are evil. In HTML they are usually not used the
        # way they should be used in LaTeX, so we replace them with spaces.
        self.assertEqual(self.convert(u'Hello\xa0World'),
                         'Hello World')

    def test_tinymce_paste(self):
        # cleanup problem with many _mcePaste divs.
        html = '\n'.join((
                '<p> </p>',
                '<div id="_mcePaste">one</div>',
                '<div id="_mcePaste"></div>',
                '<div id="_mcePaste">two</div>',
                '<div id="_mcePaste"></div>',
                '<div id="_mcePaste">three</div>',
                '<div id="_mcePaste"></div>',
                '<div id="_mcePaste">four</div>',
                '<div id="_mcePaste"></div>',
                '<div id="_mcePaste">five</div>',
                '<p>six<br />seven<br />eight</p>'))

        self.assertNotIn('div', self.convert(html))
        self.assertNotIn('mcePaste', self.convert(html))
        self.assertNotIn('<', self.convert(html))
        self.assertNotIn('>', self.convert(html))

    def test_repeat_bug(self):
        # bug was to pass re.DOTALL (which is 16) to the re.sub() function,
        # which caused to only replace the first 16 matches.
        times = 20
        html = ('<b>foo</b> ' * times).strip()
        latex = (r'\textbf{foo} ' * times).strip()

        result = self.convert(html)

        self.assertEqual(len(result), len(latex))
        self.assertEqual(result, latex)

    def test_callout_is_not_too_greedy(self):
        # Bug caused the callout pattern to be too greedy, which resulted
        # in swallowed content.
        html = '<p>foo</p> <p class="callout">bar</p> <p>baz</p>'
        result = self.convert(html)

        self.assertIn('foo', result)
        self.assertIn('bar', result)
        self.assertIn('baz', result)

    def test_no_space_after_italic(self):
        # we used to have a little space after italic text when using the
        # old school syntax {\em }. But since the new syntax \emph{} does
        # this automatically, we don't insert it manually anymore.
        html = 'foo <em>bar</em> baz'
        latex = r'foo \emph{bar} baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <u>bar</u> baz'
        latex = r'foo \emph{bar} baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <i>bar</i> baz'
        latex = r'foo \textit{bar} baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo (<em>bar</em>) baz'
        latex = r'foo (\emph{bar}) baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo (<u>bar</u>) baz'
        latex = r'foo (\emph{bar}) baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo (<i>bar</i>) baz'
        latex = r'foo (\textit{bar}) baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <em>bar</em>-baz'
        latex = r'foo \emph{bar}-baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <u>bar</u>-baz'
        latex = r'foo \emph{bar}-baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <i>bar</i>-baz'
        latex = r'foo \textit{bar}-baz'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <em>bar</em>'
        latex = r'foo \emph{bar}'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <u>bar</u>'
        latex = r'foo \emph{bar}'
        self.assertEqual(self.convert(html), latex)

        html = 'foo <i>bar</i>'
        latex = r'foo \textit{bar}'
        self.assertEqual(self.convert(html), latex)

    def test_utf8_math_charaters(self):
        self.assertEqual(self.convert(u'\u22a5'.encode('utf-8')),
                         r'$\perp$')

        self.assertEqual(self.convert(u'\u2264'.encode('utf-8')),
                         r'$\leq$')

        self.assertEqual(self.convert(u'\u2216'.encode('utf-8')),
                         r'$\setminus$')

        self.assertEqual(self.convert(u'\u2283'.encode('utf-8')),
                         r'$\supset$')

        self.assertEqual(self.convert(u'\u211c'.encode('utf-8')),
                         r'$\Re$')

        self.assertEqual(self.convert(u'\u2197'.encode('utf-8')),
                         r'$\nearrow$')

        self.assertEqual(self.convert(u'\u221a'.encode('utf-8')),
                         r'$\surd$')

        self.assertEqual(self.convert(u'\xbd'.encode('utf-8')),
                         r'$\frac12$')

        self.assertEqual(self.convert(u'\u2153'.encode('utf-8')),
                         r'$\frac13$')

        self.assertEqual(self.convert(u'\xbc'.encode('utf-8')),
                         r'$\frac14$')

        self.assertEqual(self.convert(u'\u2155'.encode('utf-8')),
                         r'$\frac15$')

        self.assertEqual(self.convert(u'\u2159'.encode('utf-8')),
                         r'$\frac16$')

        self.assertEqual(self.convert(u'\u215b'.encode('utf-8')),
                         r'$\frac18$')

        self.assertEqual(self.convert(u'\u2229'.encode('utf-8')),
                         r'$\cap$')

        self.assertEqual(self.convert(u'\u03a3'.encode('utf-8')),
                         r'$\Sigma$')

        self.assertEqual(self.convert(u'\u03a8'.encode('utf-8')),
                         r'$\Psi$')

        self.assertEqual(self.convert(u'\u03a9'.encode('utf-8')),
                         r'$\Omega$')

        self.assertEqual(self.convert(u'\u2205'.encode('utf-8')),
                         r'$\emptyset$')

        self.assertEqual(self.convert(u'\u22c3'.encode('utf-8')),
                         r'$\bigcup$')

        self.assertEqual(self.convert(u'\u2190'.encode('utf-8')),
                         r'$\leftarrow$')

        self.assertEqual(self.convert(u'\u227a'.encode('utf-8')),
                         r'$\prec$')

        self.assertEqual(self.convert(u'\u2297'.encode('utf-8')),
                         r'$\otimes$')

        self.assertEqual(self.convert(u'\u2135'.encode('utf-8')),
                         r'$\aleph$')

        self.assertEqual(self.convert(u'\u22ef'.encode('utf-8')),
                         r'$\cdots$')

        self.assertEqual(self.convert(u'\u2245'.encode('utf-8')),
                         r'$\cong$')

        self.assertEqual(self.convert(u'\u2261'.encode('utf-8')),
                         r'$\equiv$')

        self.assertEqual(self.convert(u'\u226a'.encode('utf-8')),
                         r'$\ll$')

        self.assertEqual(self.convert(u'\u22c6'.encode('utf-8')),
                         r'$\star$')

        self.assertEqual(self.convert(u'\u2260'.encode('utf-8')),
                         r'$\neq$')

        self.assertEqual(self.convert(u'\u03b1'.encode('utf-8')),
                         r'$\alpha$')

        self.assertEqual(self.convert(u'\u2210'.encode('utf-8')),
                         r'$\amalg$')

        self.assertEqual(self.convert(u'\u228e'.encode('utf-8')),
                         r'$\uplus$')

        self.assertEqual(self.convert(u'\u03ba'.encode('utf-8')),
                         r'$\kappa$')

        self.assertEqual(self.convert(u'\u03c3'.encode('utf-8')),
                         r'$\sigma$')

        self.assertEqual(self.convert(u'\u039b'.encode('utf-8')),
                         r'$\Lambda$')

        self.assertEqual(self.convert(u'\u222a'.encode('utf-8')),
                         r'$\cup$')

        self.assertEqual(self.convert(u'\u03bb'.encode('utf-8')),
                         r'$\lambda$')

        self.assertEqual(self.convert(u'\u0398'.encode('utf-8')),
                         r'$\Theta$')

        self.assertEqual(self.convert(u'\u221c'.encode('utf-8')),
                         r'$\sqrt4$')

        self.assertEqual(self.convert(u'\u2240'.encode('utf-8')),
                         r'$\wr$')

        self.assertEqual(self.convert(u'\u2118'.encode('utf-8')),
                         r'$\wp$')

        self.assertEqual(self.convert(u'\u221b'.encode('utf-8')),
                         r'$\sqrt3$')

        self.assertEqual(self.convert(u'\xac'.encode('utf-8')),
                         r'\ensuremath{\lnot}')

        self.assertEqual(self.convert(u'\u2293'.encode('utf-8')),
                         r'$\sqcap$')

        self.assertEqual(self.convert(u'\u03f1'.encode('utf-8')),
                         r'$\varrho$')

        self.assertEqual(self.convert(u'\u03b2'.encode('utf-8')),
                         r'$\beta$')

        self.assertEqual(self.convert(u'\u22a3'.encode('utf-8')),
                         r'$\dashv$')

        self.assertEqual(self.convert(u'\u2265'.encode('utf-8')),
                         r'$\geq$')

        self.assertEqual(self.convert(u'\u2199'.encode('utf-8')),
                         r'$\searrow$')

        self.assertEqual(self.convert(u'\u2664'.encode('utf-8')),
                         r'$\spadesuit$')

        self.assertEqual(self.convert(u'\u2021'.encode('utf-8')),
                         r'$\ddag$')

        self.assertEqual(self.convert(u'\u220f'.encode('utf-8')),
                         r'$\prod$')

        self.assertEqual(self.convert(u'\u2156'.encode('utf-8')),
                         r'$\frac25$')

        self.assertEqual(self.convert(u'\u2154'.encode('utf-8')),
                         r'$\frac23$')

        self.assertEqual(self.convert(u'\u22a4'.encode('utf-8')),
                         r'$\bot$')

        self.assertEqual(self.convert(u'\u21d1'.encode('utf-8')),
                         r'$\Uparrow$')

        self.assertEqual(self.convert(u'\u2194'.encode('utf-8')),
                         r'$\leftrightarrow$')

        self.assertEqual(self.convert(u'\u03c8'.encode('utf-8')),
                         r'$\psi$')

        self.assertEqual(self.convert(u'\u21c0'.encode('utf-8')),
                         r'$\rightharpoonup$')

        self.assertEqual(self.convert(u'\u2219'.encode('utf-8')),
                         r'$\bullet$')

        self.assertEqual(self.convert(u'\u03d5'.encode('utf-8')),
                         r'$\varphi$')

        self.assertEqual(self.convert(u'\u03db'.encode('utf-8')),
                         r'$\varsigma$')

        self.assertEqual(self.convert(u'\u2284'.encode('utf-8')),
                         r'$\not\subset$')

        self.assertEqual(self.convert(u'\u2206'.encode('utf-8')),
                         r'$\Delta$')

        self.assertEqual(self.convert(u'\xb0'.encode('utf-8')),
                         r'$\circ$')

        self.assertEqual(self.convert(u'\u03c5'.encode('utf-8')),
                         r'$\upsilon$')

        self.assertEqual(self.convert(u'\u03c7'.encode('utf-8')),
                         r'$\chi$')

        self.assertEqual(self.convert(u'\u2250'.encode('utf-8')),
                         r'$\doteq$')

        self.assertEqual(self.convert(u'\u2191'.encode('utf-8')),
                         r'$\uparrow$')

        self.assertEqual(self.convert(u'\u222d'.encode('utf-8')),
                         r'$\iiint$')

        self.assertEqual(self.convert(u'\u266d'.encode('utf-8')),
                         r'$\flat$')

        self.assertEqual(self.convert(u'\u03a6'.encode('utf-8')),
                         r'$\Phi$')

        self.assertEqual(self.convert(u'\u03f5'.encode('utf-8')),
                         r'$\epsilon$')

        self.assertEqual(self.convert(u'\u03d6'.encode('utf-8')),
                         r'$\varpi$')

        self.assertEqual(self.convert(u'\u22c2'.encode('utf-8')),
                         r'$\bigcap$')

        self.assertEqual(self.convert(u'\u03a0'.encode('utf-8')),
                         r'$\Pi$')

        self.assertEqual(self.convert(u'\u2020'.encode('utf-8')),
                         r'$\dag$')

        self.assertEqual(self.convert(u'\u2217'.encode('utf-8')),
                         r'$\ast$')

        self.assertEqual(self.convert(u'\u21a6'.encode('utf-8')),
                         r'$\mapsto$')

        self.assertEqual(self.convert(u'\u03b8'.encode('utf-8')),
                         r'$\theta$')

        self.assertEqual(self.convert(u'\u222e'.encode('utf-8')),
                         r'$\oint$')

        self.assertEqual(self.convert(u'\u2295'.encode('utf-8')),
                         r'$\oplus$')

        self.assertEqual(self.convert(u'\u22c4'.encode('utf-8')),
                         r'$\diamond$')

        self.assertEqual(self.convert(u'\u2113'.encode('utf-8')),
                         r'$\ell$')

        self.assertEqual(self.convert(u'\u2213'.encode('utf-8')),
                         r'$\mp$')

        self.assertEqual(self.convert(u'\u2207'.encode('utf-8')),
                         r'$\nabla$')

        self.assertEqual(self.convert(u'\u220c'.encode('utf-8')),
                         r'$\not\ni$')

        self.assertEqual(self.convert(u'\u2298'.encode('utf-8')),
                         r'$\oslash$')

        self.assertEqual(self.convert(u'\u222c'.encode('utf-8')),
                         r'$\iint$')

        self.assertEqual(self.convert(u'\u22c5'.encode('utf-8')),
                         r'$\cdot$')

        self.assertEqual(self.convert(u'\u2111'.encode('utf-8')),
                         r'$\Im$')

        self.assertEqual(self.convert(u'\xb1'.encode('utf-8')),
                         r'$\pm$')

        self.assertEqual(self.convert(u'\u03c0'.encode('utf-8')),
                         r'$\pi$')

        self.assertEqual(self.convert(u'\u2287'.encode('utf-8')),
                         r'$\supseteq$')

        self.assertEqual(self.convert(u'\u0393'.encode('utf-8')),
                         r'$\Gamma$')

        self.assertEqual(self.convert(u'\u03c4'.encode('utf-8')),
                         r'$\tau$')

        self.assertEqual(self.convert(u'\u2299'.encode('utf-8')),
                         r'$\odot$')

        self.assertEqual(self.convert(u'\u21aa'.encode('utf-8')),
                         r'$\hookrightarrow$')

        self.assertEqual(self.convert(u'\u2294'.encode('utf-8')),
                         r'$\sqcup$')

        self.assertEqual(self.convert(u'\u22a8'.encode('utf-8')),
                         r'$\models$')

        self.assertEqual(self.convert(u'\u21d2'.encode('utf-8')),
                         r'$\Rightarrow$')

        self.assertEqual(self.convert(u'\u22c0'.encode('utf-8')),
                         r'$\bigwedge$')

        self.assertEqual(self.convert(u'\u03c9'.encode('utf-8')),
                         r'$\omega$')

        self.assertEqual(self.convert(u'\u215c'.encode('utf-8')),
                         r'$\frac38$')

        self.assertEqual(self.convert(u'\u221d'.encode('utf-8')),
                         r'$\propto$')

        self.assertEqual(self.convert(u'\u266f'.encode('utf-8')),
                         r'$\sharp$')

        self.assertEqual(self.convert(u'\u03bc'.encode('utf-8')),
                         r'$\mu$')

        self.assertEqual(self.convert(u'\u2157'.encode('utf-8')),
                         r'$\frac35$')

        self.assertEqual(self.convert(u'\xbe'.encode('utf-8')),
                         r'$\frac34$')

        self.assertEqual(self.convert(u'\u220a'.encode('utf-8')),
                         r'$\in$')

        self.assertEqual(self.convert(u'\u03b5'.encode('utf-8')),
                         r'$\varepsilon$')

        self.assertEqual(self.convert(u'\u03b6'.encode('utf-8')),
                         r'$\zeta$')

        self.assertEqual(self.convert(u'\u03b7'.encode('utf-8')),
                         r'$\eta$')

        self.assertEqual(self.convert(u'\u2244'.encode('utf-8')),
                         r'$\not\simeq$')

        self.assertEqual(self.convert(u'\u2158'.encode('utf-8')),
                         r'$\frac45$')

        self.assertEqual(self.convert(u'\u2200'.encode('utf-8')),
                         r'$\forall$')

        self.assertEqual(self.convert(u'\xd7'.encode('utf-8')),
                         r'$\times$')

        self.assertEqual(self.convert(u'\u2026'.encode('utf-8')),
                         r'$\dots$')

        self.assertEqual(self.convert(u'\u2262'.encode('utf-8')),
                         r'$\not\equiv$')

        self.assertEqual(self.convert(u'\u2196'.encode('utf-8')),
                         r'$\nwarrow$')

        self.assertEqual(self.convert(u'\u2227'.encode('utf-8')),
                         r'$\wedge$')

        self.assertEqual(self.convert(u'\u03c1'.encode('utf-8')),
                         r'$\rho$')

        self.assertEqual(self.convert(u'\u2666'.encode('utf-8')),
                         r'$\diamondsuit$')

        self.assertEqual(self.convert(u'\u220d'.encode('utf-8')),
                         r'$\ni$')

        self.assertEqual(self.convert(u'\u2192'.encode('utf-8')),
                         r'$\rightarrow$')

        self.assertEqual(self.convert(u'\u039e'.encode('utf-8')),
                         r'$\Xi$')

        self.assertEqual(self.convert(u'\u03b9'.encode('utf-8')),
                         r'$\iota$')

        self.assertEqual(self.convert(u'\u03bd'.encode('utf-8')),
                         r'$\nu$')

        self.assertEqual(self.convert(u'\xf7'.encode('utf-8')),
                         r'$\div$')

        self.assertEqual(self.convert(u'\xa7'.encode('utf-8')),
                         r'\S')

        self.assertEqual(self.convert(u'\u03d1'.encode('utf-8')),
                         r'$\vartheta$')

        self.assertEqual(self.convert(u'\u03c6'.encode('utf-8')),
                         r'$\phi$')

        self.assertEqual(self.convert(u'\u22a2'.encode('utf-8')),
                         r'$\vdash$')

        self.assertEqual(self.convert(u'\u2209'.encode('utf-8')),
                         r'$\notin$')

        self.assertEqual(self.convert(u'\u2296'.encode('utf-8')),
                         r'$\ominus$')

        self.assertEqual(self.convert(u'\u210f'.encode('utf-8')),
                         r'$\hbar$')

        self.assertEqual(self.convert(u'\u2220'.encode('utf-8')),
                         r'$\angle$')

        self.assertEqual(self.convert(u'\u227b'.encode('utf-8')),
                         r'$\succ$')

        self.assertEqual(self.convert(u'\u221e'.encode('utf-8')),
                         r'$\infty$')

        self.assertEqual(self.convert(u'\u03be'.encode('utf-8')),
                         r'$\xi$')

        self.assertEqual(self.convert(u'\u2665'.encode('utf-8')),
                         r'$\heartsuit$')

        self.assertEqual(self.convert(u'\u2249'.encode('utf-8')),
                         r'$\not\approx$')

        self.assertEqual(self.convert(u'\xb6'.encode('utf-8')),
                         r'\P')

        self.assertEqual(self.convert(u'\u21d0'.encode('utf-8')),
                         r'$\Leftarrow$')

        self.assertEqual(self.convert(u'\u2228'.encode('utf-8')),
                         r'$\vee$')

        self.assertEqual(self.convert(u'\u03b3'.encode('utf-8')),
                         r'$\gamma$')

        self.assertEqual(self.convert(u'\u2195'.encode('utf-8')),
                         r'$\updownarrow$')

        self.assertEqual(self.convert(u'\u224d'.encode('utf-8')),
                         r'$\asymp$')

        self.assertEqual(self.convert(u'\u222b'.encode('utf-8')),
                         r'$\int$')

        self.assertEqual(self.convert(u'\u223c'.encode('utf-8')),
                         r'$\sim$')

        self.assertEqual(self.convert(u'\u2667'.encode('utf-8')),
                         r'$\clubsuit$')

        self.assertEqual(self.convert(u'\u215e'.encode('utf-8')),
                         r'$\frac78$')

        self.assertEqual(self.convert(u'\u21a9'.encode('utf-8')),
                         r'$\hookleftarrow$')

        self.assertEqual(self.convert(u'\u2193'.encode('utf-8')),
                         r'$\downarrow$')

        self.assertEqual(self.convert(u'\u226b'.encode('utf-8')),
                         r'$\gg$')

        self.assertEqual(self.convert(u'\u22c1'.encode('utf-8')),
                         r'$\bigvee$')

        self.assertEqual(self.convert(u'\u226d'.encode('utf-8')),
                         r'$\not\asymp$')

        self.assertEqual(self.convert(u'\u21d5'.encode('utf-8')),
                         r'$\Updownarrow$')

        self.assertEqual(self.convert(u'\u03b4'.encode('utf-8')),
                         r'$\delta$')

        self.assertEqual(self.convert(u'\u21bc'.encode('utf-8')),
                         r'$\leftharpoonup$')

        self.assertEqual(self.convert(u'\u2243'.encode('utf-8')),
                         r'$\simeq$')

        self.assertEqual(self.convert(u'\u2225'.encode('utf-8')),
                         r'$\parallel$')

        self.assertEqual(self.convert(u'\u215d'.encode('utf-8')),
                         r'$\frac58$')

        self.assertEqual(self.convert(u'\u215a'.encode('utf-8')),
                         r'$\frac56$')

        self.assertEqual(self.convert(u'\u2248'.encode('utf-8')),
                         r'$\approx$')

        self.assertEqual(self.convert(u'\u21c1'.encode('utf-8')),
                         r'$\rightharpoondown$')

        self.assertEqual(self.convert(u'\u2202'.encode('utf-8')),
                         r'$\partial$')

        self.assertEqual(self.convert(u'\u2286'.encode('utf-8')),
                         r'$\subseteq$')

        self.assertEqual(self.convert(u'\u21cc'.encode('utf-8')),
                         r'$\rightleftharpoons$')

        self.assertEqual(self.convert(u'\u2223'.encode('utf-8')),
                         r'$\mid$')

        self.assertEqual(self.convert(u'\u215f'.encode('utf-8')),
                         r'$\frac14$')

        self.assertEqual(self.convert(u'\u21bd'.encode('utf-8')),
                         r'$\leftharpoondown$')

        self.assertEqual(self.convert(u'\u21d3'.encode('utf-8')),
                         r'$\Downarrow$')

        self.assertEqual(self.convert(u'\u22c8'.encode('utf-8')),
                         r'$\bowtie$')

        self.assertEqual(self.convert(u'\u266e'.encode('utf-8')),
                         r'$\natural$')

        self.assertEqual(self.convert(u'\u22ee'.encode('utf-8')),
                         r'$\vdots$')

        self.assertEqual(self.convert(u'\u2211'.encode('utf-8')),
                         r'$\sum$')

        self.assertEqual(self.convert(u'\u2285'.encode('utf-8')),
                         r'$\not\supset$')

        self.assertEqual(self.convert(u'\u2282'.encode('utf-8')),
                         r'$\subset$')

    def test_swiss_currency_with_dash(self):
        self.assertEqual([
                'Fr.~0.--',
                'Fr.~123.--',
                'Fr.~45.--',
                'Fr.~67.--',
                'Fr. .-',

                ], [
                self.convert('Fr. 0.-'),
                self.convert('Fr. 123.-'),
                self.convert('Fr. 45.--'),
                self.convert('Fr. 67.\xe2\x80\x93'),
                self.convert('Fr. .-'),
                ])

    def test_html_comments_are_removed(self):
        samples = (
            ('<!-- foo -->', ''),
            ('<!-- foo --> bar <!-- baz -->', 'bar'),
            ('<!-- <b>foo</b> -->', ''),
            ('<!--foo-->', ''),
            ('<!--[foo]-->', ''),
            ('<!--[if x]> <b>bar</b> <![endif]-->', ''),
            )

        self.assertEquals(
            [item[1] for item in samples],
            [self.convert(item[0]).strip() for item in samples])

    def test_cleans_up_bad_word_comments(self):
        input = '\n'.join((
                '<!--[if gte mso 9]><xml>',
                '<o:OfficeDocumentSettings>',
                '<o:TargetScreenSize>800x600</o:TargetScreenSize>',
                '</o:OfficeDocumentSettings>',
                '</xml><![endif]-->',

                ))

        self.assertEqual('', self.convert(input).strip())

    def test_hyphenation_infos_are_added_to_urls(self):
        self.assertEqual(
            self.convert(
                'x http://foo.ch/bar y'),
            'x http://foo.ch""/bar y')

        self.assertEqual(
            self.convert(
                'x https://user@pass:domain.com/foo/bar.html?param=true#anchor y'),
            'x https://user@pass:domain.com""/foo""/bar.html?param=true\\#anchor y')

    def assert_convertions(self, input_expected):
        input_output = {}

        for input, _expected in input_expected.items():
            input_output[input] = self.convert(input).strip()

        self.assertEqual(input_expected, input_output)



"""Unit tests testing the default configuration utility.
"""

from ftw.pdfgenerator.config import DefaultConfig
from ftw.pdfgenerator.interfaces import IConfig
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from unittest2 import TestCase
from zope.component import queryUtility
from zope.interface.verify import verifyClass
import os


class TestDefaultConfigurationUtility(TestCase):

    layer = PDFGENERATOR_ZCML_LAYER

    def test_config_implements_interface(self):
        self.assertTrue(IConfig.implementedBy(DefaultConfig))
        verifyClass(IConfig, DefaultConfig)

    def test_remove_build_directory_true_by_default(self):
        self.assertTrue(DefaultConfig().remove_build_directory)

    def test_build_directory_existing(self):
        path = DefaultConfig().get_build_directory()
        self.assertTrue(os.path.exists(path))
        os.rmdir(path)

    def test_build_directory_is_writeable(self):
        path = DefaultConfig().get_build_directory()
        self.assertTrue(os.access(path, os.R_OK | os.W_OK))
        os.rmdir(path)

    def test_build_directory_not_the_same_twice(self):
        config = DefaultConfig()
        path_one = config.get_build_directory()
        path_two = config.get_build_directory()

        self.assertNotEqual(path_one, path_two)
        os.rmdir(path_one)
        os.rmdir(path_two)

    def test_config_utility_is_registered_and_default_utility(self):
        self.assertIsNotNone(queryUtility(IConfig))

        self.assertEqual(queryUtility(IConfig).__class__, DefaultConfig)



from ftw.pdfgenerator import interfaces
from ftw.pdfgenerator.html2latex import subconverter
from ftw.pdfgenerator.utils import decode_htmlentities


class HtmlentitiesConverter(subconverter.SubConverter):

    pattern = r'\\&(\w*);'
    placeholder = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_BOTTOM

    def __call__(self):
        html = self.get_html()[1:]
        latex = decode_htmlentities(html).encode('utf8')
        self.replace(latex)



from ftw.pdfgenerator.html2latex.utils import generate_manual_caption
from ftw.pdfgenerator.testing import ZCML_WITH_SITE_LAYER
from ftw.testing import MockTestCase


class TestGenerateManualCaption(MockTestCase):

    layer = ZCML_WITH_SITE_LAYER

    def test_empty_caption(self):
        self.assertEqual(generate_manual_caption('', 'table'), '')
        self.assertEqual(generate_manual_caption(None, 'table'), '')

    def test_wrong_index(self):
        with self.assertRaises(ValueError) as cm:
            generate_manual_caption('foo', 'wrong')

        self.assertEqual(
            str(cm.exception),
            'Environment must be one of "table", "figure". Got "wrong".')

    def test_table_caption(self):
        latex = generate_manual_caption('The Caption', 'table')
        self.assertIn(r'\addtocounter{table}{1}', latex)
        self.assertIn(r'\addcontentsline{lot}{table}', latex)
        self.assertIn(r'\ignorespaces The Caption', latex)
        self.assertIn(r'Table \thechapter.\arabic{table}: The Caption',
                      latex)

    def test_figure_caption(self):
        latex = generate_manual_caption('The Caption', 'figure')
        self.assertIn(r'\addtocounter{figure}{1}', latex)
        self.assertIn(r'\addcontentsline{lof}{figure}', latex)
        self.assertIn(r'\ignorespaces The Caption', latex)
        self.assertIn(r'Figure \thechapter.\arabic{figure}: The Caption',
                      latex)

    def test_not_listed_table_caption(self):
        latex = generate_manual_caption('The Caption', 'table',
                                        show_in_index=False)

        self.assertNotIn(r'\addtocounter{table}{1}', latex)
        self.assertNotIn(r'\addcontentsline{lot}{table}', latex)
        self.assertNotIn(r'\ignorespaces The Caption', latex)
        self.assertNotIn(r'Table \thechapter.\arabic{table}: The Caption',
                      latex)
        self.assertIn(r'The Caption',
                      latex)

    def test_caption_with_umlaut(self):
        latex = generate_manual_caption('hall\xc3\xb6chen', 'table')
        self.assertIn(
            'Table \\thechapter.\\arabic{table}: hall\xc3\xb6chen',
            latex)



from ftw.pdfgenerator import interfaces
from ftw.pdfgenerator.html2latex import subconverter
import re

def find_bracket_range(latex):
    opening = -1
    brackets = 0
    xpr = re.compile('[\{\}]')

    for match in xpr.finditer(latex):
        if match.group() == '{':
            if opening == -1:
                opening = match.span()[0]
            brackets += 1

        elif match.group() == '}':
            if opening != -1:
                brackets -= 1

        if opening != -1 and brackets == 0:
            return opening, match.span()[1]

    return None


class Textformatting(subconverter.SubConverter):
    """ This converter fixes newlines within text formatting commands:

    "\textbf{*\n\n*}" is not allowed, use "\textbf{*\\*}"
    "\textit{*\n\n*}" is not allowed, use "\textit{*\\*}"
    "\emph{*\n\n*}" is not allowed, use "\emph{*\\*}"
    """

    pattern = r'(\\(textbf|textit|emph){.*})'
    placeholder = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_BOTTOM

    def __call__(self):
        latex = self.get_html()
        head, tail = find_bracket_range(latex)

        before = latex[:head]
        environment = latex[head:tail]
        after = latex[tail:]

        environment = self.convert_substring(environment)

        latex = before + environment + after
        self.replace(latex)

    def convert_substring(self, latex):
        return latex.replace('\n\n', r'\\\\')



# -*- coding: latin-1 -*-

from ftw.pdfgenerator import interfaces


MODE_REPLACE = interfaces.HTML2LATEX_MODE_REPLACE
MODE_REGEXP = interfaces.HTML2LATEX_MODE_REGEXP
MODE_REGEXP_FUNCTION = interfaces.HTML2LATEX_MODE_REGEXP_FUNCTION
BACKSLASH_MARKER = 'THISISABACKSLASH' * 2

DASH = '\xe2\x80\x93'  # &ndash;


DEFAULT_PATTERNS = ([
        (MODE_REPLACE,  '\\',                      BACKSLASH_MARKER),
        # remove ASCII non breaking space:
        (MODE_REPLACE,  '\xc2\xa0',                ' '),
        # remove ASCII vertical tab control character
        (MODE_REPLACE,  '\x0b',                ' '),
        # remove ASCII form feed control character
        (MODE_REPLACE,  '\x0c',                ' '),
        interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_TOP,

        # special characters
        (MODE_REPLACE,  '{',                       '\\{'),
        (MODE_REPLACE,  '}',                       '\\}'),
        interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER,
        (MODE_REPLACE,  '&nbsp;',                  ' '),
        (MODE_REGEXP,   '',                    '--'),
        (MODE_REGEXP,   '',                      ''),

        # remove comments
        (MODE_REGEXP,  r'<!--(?:(?!-->).)*-->',    ''),

        # generally remove empty no-singleton-tags
        (MODE_REGEXP,  r'<(?P<tag>[\w]{0,})>[\s]{0,}</(?P=tag)?>',
         r'',         (interfaces.HTML2LATEX_REPEAT_MODIFIER)),

        # we use only \n for carriage return, remove all \r
        (MODE_REPLACE,  '\r\n',                    ' '),
        (MODE_REPLACE,  '\r',                      ' '),
        (MODE_REPLACE,  '\n',                      ' '),

        # remove "blacklisted" tags (with its contents)
        (MODE_REGEXP,   r'<style.*?>.*?</style>',   r''),

        # headings
        (MODE_REGEXP,   r'<span>(.*?)</span>',     r'\g<1>'),
        (MODE_REGEXP,   r'<h1.*?>(.*?)</h1>',      r'\\section{\g<1>}\n'),
        (MODE_REGEXP,   r'<h2.*?>(.*?)</h2>',      r'\\subsection{\g<1>}\n'),
        (MODE_REGEXP,   r'<h3.*?>(.*?)</h3>',
         r'\\subsubsection{\g<1>}\n'),

        (MODE_REGEXP,   r'<h5.*?class="glossar">(.*?)</h5>',
         r'\\minisec{\g<1>}\n'),
        (MODE_REGEXP,   r'<span.*?class="paragraphHeading".*?>(.*?)</span>',
         r'\\subparagraph{\g<1>}\n'),

        # font style and text decoration
        (MODE_REGEXP,   r'<p[^>]*?class="callout"[^>]*?>(.*?)</p>',
         r'\\begin{quote}\n\g<1>\n\\end{quote}\n'),
        (MODE_REGEXP,   r'([^ ])<(b|strong|em|u|i)> ',
         r'\g<1> <\g<2>>'),
        (MODE_REGEXP,   r'<b>(.*?)</b>',
         r'\\textbf{\g<1>}' + interfaces.HTML2LATEX_PREVENT_CHARACTER),
        (MODE_REGEXP,   r'<b .*?>(.*?)</b>',
         r'\\textbf{\g<1>}' + interfaces.HTML2LATEX_PREVENT_CHARACTER),
        (MODE_REGEXP,   r'<strong.*?>(.*?)</strong>',
         r'\\textbf{\g<1>}' + interfaces.HTML2LATEX_PREVENT_CHARACTER),

        (MODE_REGEXP,   r'<span .*?style="font-weight: bold.*?">(.*?)</span>',
         r'\\textbf{\g<1>}' + interfaces.HTML2LATEX_PREVENT_CHARACTER),

        # <em> / <u> -> \emph{X}
        (MODE_REGEXP,   r'<(em|u)(>| [^>]*>)(.*?)</(\1)>',
         r'\\emph{\g<3>}'),
        # <i> -> \textit{X}
        (MODE_REGEXP,   r'<(i)(>| [^>]*>)(.*?)</(\1)>',
         r'\\textit{\g<3>}'),

        (MODE_REGEXP,   r'<sup.*?>(.*?)</sup>',
         r'\\textsuperscript{\g<1>}'),

        # quotes
        (MODE_REGEXP,   r'\&quot;',                 '"'),
        (MODE_REGEXP,   r'"([\w\{\\])',             '"`\g<1>'),
        (MODE_REGEXP,   r'([^ \t\n\r\f\v\{])"(?!`)',     '\g<1>"\''),

        # paragraphs and white space
        (MODE_REGEXP,   r'<p.*?>(.*?)[\r\n ]{0,}</p>',
         r'\g<1>\n\n'),
        (MODE_REGEXP,   r'[\r\n]{0,}<br[ \W]{0,}>\s*<br[ \W]{0,}>[\n]{0,1}',
         r'\n\n'),
        (MODE_REGEXP,   r'[\r\n]{0,}<br[ \W]{0,}>\n',
         r'\n'),
        (MODE_REGEXP,   r'[\r\n]{0,}<br[ \W]{0,}>',
         r'\n'),

        (MODE_REGEXP,   r'[\n]{2,}',               r'\n\n'),
        (MODE_REGEXP,   r'^\s*(.*?)\s*$',          r'\g<1>'),           # trim
        (MODE_REGEXP,   r'\n([^\n])',              r'\\\\\n\g<1>'),
        (MODE_REGEXP,   r'([^}])\n\n',             r'\g<1>\\\\\n\n'),
        (MODE_REGEXP,   r'({\\[\w]*? )\\\\\W(.*?})',
         r'\1\2'),
        (MODE_REGEXP,   r'(\\[\w]*?{)\\\\\W(.*?})',
         r'\1\2'),


        # special characters
        (MODE_REPLACE,  '%',                       '\\%'),
        (MODE_REGEXP,   r'(\d) \\%',               r'\1\,\\%'),
        (MODE_REPLACE,  '_',                       '\\_'),
        (MODE_REPLACE,  '$',                       '\\$'),
        (MODE_REGEXP,    '§',                       '\\S'),
        (MODE_REPLACE,  '#',                       '\\#'),
        (MODE_REGEXP,    '&szlig;',                 'ß'),
        (MODE_REPLACE,  '&euro;',                  '\\euro{}'),
        (MODE_REPLACE,  '&sect;',                  '\\S'),
        (MODE_REPLACE,  '&amp;',                   '&'),
        (MODE_REPLACE,  '*',                       r'$\ast$'),

        (MODE_REPLACE,  '&lsquo;',                 '‘'),
        (MODE_REPLACE,  '&rsquo;',                 '’'),
        (MODE_REPLACE,  '&rsquo;',                 '’'),
        (MODE_REPLACE,  '&ldquo;',                 '“'),
        (MODE_REPLACE,  '&rdquo;',                 '”'),
        (MODE_REPLACE,  '&ndash;',                 '–'),
        (MODE_REPLACE,  '&mdash;',                 '—'),
        (MODE_REPLACE,  '&iexcl;',                 '¡'),
        (MODE_REPLACE,  '&cent;',                  '\\cent{}'),
        (MODE_REPLACE,  '&pound;',                 '£'),
        (MODE_REPLACE,  '&curren;',                '\\currency{}'),
        (MODE_REPLACE,  '&sect;',                  '§'),

        (MODE_REPLACE,  '&uml;',                   '¨'),
        (MODE_REPLACE,  '&copy;',                  '©'),
        (MODE_REPLACE,  '&ordf;',                  'ª'),
        (MODE_REPLACE,  '&laquo;',                 '«'),
        (MODE_REPLACE,  '&not;',                   r'\ensuremath{\lnot}'),
        (MODE_REPLACE,  '&reg;',                   '®'),
        (MODE_REPLACE,  '&macr;',                  '¯'),
        (MODE_REPLACE,  '&deg;',                   '°'),
        (MODE_REPLACE,  '&plusmn;',                '±'),

        (MODE_REPLACE,  '&sup2;',                  '²'),
        (MODE_REPLACE,  '&sup3;',                  '³'),
        (MODE_REPLACE,  '&acute;',                 '´'),
        (MODE_REPLACE,  '&micro;',                 'µ'),
        (MODE_REPLACE,  '&para;',                  '¶'),
        (MODE_REPLACE,  '&middot;',                '·'),
        (MODE_REPLACE,  '&cedil;',                 '¸'),

        (MODE_REPLACE,  '&sup1;',                  '¹'),
        (MODE_REPLACE,  '&ordm;',                  'º'),
        (MODE_REPLACE,  '&raquo;',                 '»'),
        (MODE_REPLACE,  '&frac14;',                '¼'),
        (MODE_REPLACE,  '&frac12;',                '½'),
        (MODE_REPLACE,  '&frac34;',                '¾'),

        (MODE_REPLACE,  '&iquest;',                '¿'),
        (MODE_REPLACE,  '&times;',                 '×'),
        (MODE_REPLACE,  '&divide;',                '÷'),
        (MODE_REPLACE,  '&sbquo;',                 '‚'),
        (MODE_REPLACE,  '&bdquo;',                 '„'),
        (MODE_REPLACE,  '&hellip;',                '…'),

        (MODE_REPLACE,  '&trade;',                 '™'),
        (MODE_REPLACE,  '&bull;',                  '•'),
        (MODE_REPLACE,  '&rarr;',                  '→'),
        (MODE_REPLACE,  '&rArr;',                  '⇒'),
        (MODE_REPLACE,  '&hArr;',                  '⇔'),
        (MODE_REPLACE,  '&asymp;',                 '≈'),
        (MODE_REPLACE,  '‑',                       '-'),
        (MODE_REPLACE,  '^',                       '\\^{}'),
        (MODE_REPLACE,  '~',                       '\\~{}'),

        # additional hypthenation character
        (MODE_REPLACE,  '\xc2\xad',                '"-'),
        (MODE_REPLACE,  '&shy;',                   '"-'),

        # utf8 math characters
        (MODE_REPLACE,  '\xe2\x8a\xa5',            r'$\perp$'),
        (MODE_REPLACE,  '\xe2\x89\xa4',            r'$\leq$'),
        (MODE_REPLACE,  '\xe2\x88\x96',            r'$\setminus$'),
        (MODE_REPLACE,  '\xe2\x8a\x83',            r'$\supset$'),
        (MODE_REPLACE,  '\xe2\x84\x9c',            r'$\Re$'),
        (MODE_REPLACE,  '\xe2\x86\x97',            r'$\nearrow$'),
        (MODE_REPLACE,  '\xe2\x88\x9a',            r'$\surd$'),
        (MODE_REPLACE,  '\xc2\xbd',                r'$\frac12$'),
        (MODE_REPLACE,  '\xe2\x85\x93',            r'$\frac13$'),
        (MODE_REPLACE,  '\xc2\xbc',                r'$\frac14$'),
        (MODE_REPLACE,  '\xe2\x85\x95',            r'$\frac15$'),
        (MODE_REPLACE,  '\xe2\x85\x99',            r'$\frac16$'),
        (MODE_REPLACE,  '\xe2\x85\x9b',            r'$\frac18$'),
        (MODE_REPLACE,  '\xe2\x88\xa9',            r'$\cap$'),
        (MODE_REPLACE,  '\xce\xa3',                r'$\Sigma$'),
        (MODE_REPLACE,  '\xce\xa8',                r'$\Psi$'),
        (MODE_REPLACE,  '\xce\xa9',                r'$\Omega$'),
        (MODE_REPLACE,  '\xe2\x88\x85',            r'$\emptyset$'),
        (MODE_REPLACE,  '\xe2\x8b\x83',            r'$\bigcup$'),
        (MODE_REPLACE,  '\xe2\x86\x90',            r'$\leftarrow$'),
        (MODE_REPLACE,  '\xe2\x89\xba',            r'$\prec$'),
        (MODE_REPLACE,  '\xe2\x8a\x97',            r'$\otimes$'),
        (MODE_REPLACE,  '\xe2\x84\xb5',            r'$\aleph$'),
        (MODE_REPLACE,  '\xe2\x8b\xaf',            r'$\cdots$'),
        (MODE_REPLACE,  '\xe2\x89\x85',            r'$\cong$'),
        (MODE_REPLACE,  '\xe2\x89\xa1',            r'$\equiv$'),
        (MODE_REPLACE,  '\xe2\x89\xaa',            r'$\ll$'),
        (MODE_REPLACE,  '\xe2\x8b\x86',            r'$\star$'),
        (MODE_REPLACE,  '\xe2\x89\xa0',            r'$\neq$'),
        (MODE_REPLACE,  '\xce\xb1',                r'$\alpha$'),
        (MODE_REPLACE,  '\xe2\x88\x90',            r'$\amalg$'),
        (MODE_REPLACE,  '\xe2\x8a\x8e',            r'$\uplus$'),
        (MODE_REPLACE,  '\xce\xba',                r'$\kappa$'),
        (MODE_REPLACE,  '\xcf\x83',                r'$\sigma$'),
        (MODE_REPLACE,  '\xce\x9b',                r'$\Lambda$'),
        (MODE_REPLACE,  '\xe2\x88\xaa',            r'$\cup$'),
        (MODE_REPLACE,  '\xce\xbb',                r'$\lambda$'),
        (MODE_REPLACE,  '\xce\x98',                r'$\Theta$'),
        (MODE_REPLACE,  '\xe2\x88\x9c',            r'$\sqrt4$'),
        (MODE_REPLACE,  '\xe2\x89\x80',            r'$\wr$'),
        (MODE_REPLACE,  '\xe2\x84\x98',            r'$\wp$'),
        (MODE_REPLACE,  '\xe2\x88\x9b',            r'$\sqrt3$'),
        (MODE_REPLACE,  '\xc2\xac',                r'\ensuremath{\lnot}'),
        (MODE_REPLACE,  '\xe2\x8a\x93',            r'$\sqcap$'),
        (MODE_REPLACE,  '\xcf\xb1',                r'$\varrho$'),
        (MODE_REPLACE,  '\xce\xb2',                r'$\beta$'),
        (MODE_REPLACE,  '\xe2\x8a\xa3',            r'$\dashv$'),
        (MODE_REPLACE,  '\xe2\x89\xa5',            r'$\geq$'),
        (MODE_REPLACE,  '\xe2\x86\x99',            r'$\searrow$'),
        (MODE_REPLACE,  '\xe2\x99\xa4',            r'$\spadesuit$'),
        (MODE_REPLACE,  '\xe2\x80\xa1',            r'$\ddag$'),
        (MODE_REPLACE,  '\xe2\x88\x8f',            r'$\prod$'),
        (MODE_REPLACE,  '\xe2\x85\x96',            r'$\frac25$'),
        (MODE_REPLACE,  '\xe2\x85\x94',            r'$\frac23$'),
        (MODE_REPLACE,  '\xe2\x8a\xa4',            r'$\bot$'),
        (MODE_REPLACE,  '\xe2\x87\x91',            r'$\Uparrow$'),
        (MODE_REPLACE,  '\xe2\x86\x94',            r'$\leftrightarrow$'),
        (MODE_REPLACE,  '\xcf\x88',                r'$\psi$'),
        (MODE_REPLACE,  '\xe2\x87\x80',            r'$\rightharpoonup$'),
        (MODE_REPLACE,  '\xe2\x88\x99',            r'$\bullet$'),
        (MODE_REPLACE,  '\xcf\x95',                r'$\varphi$'),
        (MODE_REPLACE,  '\xcf\x9b',                r'$\varsigma$'),
        (MODE_REPLACE,  '\xe2\x8a\x84',            r'$\not\subset$'),
        (MODE_REPLACE,  '\xe2\x88\x86',            r'$\Delta$'),
        (MODE_REPLACE,  '\xc2\xb0',                r'$\circ$'),
        (MODE_REPLACE,  '\xcf\x85',                r'$\upsilon$'),
        (MODE_REPLACE,  '\xcf\x87',                r'$\chi$'),
        (MODE_REPLACE,  '\xe2\x89\x90',            r'$\doteq$'),
        (MODE_REPLACE,  '\xe2\x86\x91',            r'$\uparrow$'),
        (MODE_REPLACE,  '\xe2\x88\xad',            r'$\iiint$'),
        (MODE_REPLACE,  '\xe2\x99\xad',            r'$\flat$'),
        (MODE_REPLACE,  '\xce\xa6',                r'$\Phi$'),
        (MODE_REPLACE,  '\xcf\xb5',                r'$\epsilon$'),
        (MODE_REPLACE,  '\xcf\x96',                r'$\varpi$'),
        (MODE_REPLACE,  '\xe2\x8b\x82',            r'$\bigcap$'),
        (MODE_REPLACE,  '\xce\xa0',                r'$\Pi$'),
        (MODE_REPLACE,  '\xe2\x80\xa0',            r'$\dag$'),
        (MODE_REPLACE,  '\xe2\x88\x97',            r'$\ast$'),
        (MODE_REPLACE,  '\xe2\x86\xa6',            r'$\mapsto$'),
        (MODE_REPLACE,  '\xce\xb8',                r'$\theta$'),
        (MODE_REPLACE,  '\xe2\x88\xae',            r'$\oint$'),
        (MODE_REPLACE,  '\xe2\x8a\x95',            r'$\oplus$'),
        (MODE_REPLACE,  '\xe2\x8b\x84',            r'$\diamond$'),
        (MODE_REPLACE,  '\xe2\x84\x93',            r'$\ell$'),
        (MODE_REPLACE,  '\xe2\x88\x93',            r'$\mp$'),
        (MODE_REPLACE,  '\xe2\x88\x87',            r'$\nabla$'),
        (MODE_REPLACE,  '\xe2\x88\x8c',            r'$\not\ni$'),
        (MODE_REPLACE,  '\xe2\x8a\x98',            r'$\oslash$'),
        (MODE_REPLACE,  '\xe2\x88\xac',            r'$\iint$'),
        (MODE_REPLACE,  '\xe2\x8b\x85',            r'$\cdot$'),
        (MODE_REPLACE,  '\xe2\x84\x91',            r'$\Im$'),
        (MODE_REPLACE,  '\xc2\xb1',                r'$\pm$'),
        (MODE_REPLACE,  '\xcf\x80',                r'$\pi$'),
        (MODE_REPLACE,  '\xe2\x8a\x87',            r'$\supseteq$'),
        (MODE_REPLACE,  '\xce\x93',                r'$\Gamma$'),
        (MODE_REPLACE,  '\xcf\x84',                r'$\tau$'),
        (MODE_REPLACE,  '\xe2\x8a\x99',            r'$\odot$'),
        (MODE_REPLACE,  '\xe2\x86\xaa',            r'$\hookrightarrow$'),
        (MODE_REPLACE,  '\xe2\x8a\x94',            r'$\sqcup$'),
        (MODE_REPLACE,  '\xe2\x8a\xa8',            r'$\models$'),
        (MODE_REPLACE,  '\xe2\x87\x92',            r'$\Rightarrow$'),
        (MODE_REPLACE,  '\xe2\x8b\x80',            r'$\bigwedge$'),
        (MODE_REPLACE,  '\xcf\x89',                r'$\omega$'),
        (MODE_REPLACE,  '\xe2\x85\x9c',            r'$\frac38$'),
        (MODE_REPLACE,  '\xe2\x88\x9d',            r'$\propto$'),
        (MODE_REPLACE,  '\xe2\x99\xaf',            r'$\sharp$'),
        (MODE_REPLACE,  '\xce\xbc',                r'$\mu$'),
        (MODE_REPLACE,  '\xe2\x85\x97',            r'$\frac35$'),
        (MODE_REPLACE,  '\xc2\xbe',                r'$\frac34$'),
        (MODE_REPLACE,  '\xe2\x88\x8a',            r'$\in$'),
        (MODE_REPLACE,  '\xce\xb5',                r'$\varepsilon$'),
        (MODE_REPLACE,  '\xce\xb6',                r'$\zeta$'),
        (MODE_REPLACE,  '\xce\xb7',                r'$\eta$'),
        (MODE_REPLACE,  '\xe2\x89\x84',            r'$\not\simeq$'),
        (MODE_REPLACE,  '\xe2\x85\x98',            r'$\frac45$'),
        (MODE_REPLACE,  '\xe2\x88\x80',            r'$\forall$'),
        (MODE_REPLACE,  '\xc3\x97',                r'$\times$'),
        (MODE_REPLACE,  '\xe2\x80\xa6',            r'$\dots$'),
        (MODE_REPLACE,  '\xe2\x89\xa2',            r'$\not\equiv$'),
        (MODE_REPLACE,  '\xe2\x86\x96',            r'$\nwarrow$'),
        (MODE_REPLACE,  '\xe2\x88\xa7',            r'$\wedge$'),
        (MODE_REPLACE,  '\xcf\x81',                r'$\rho$'),
        (MODE_REPLACE,  '\xe2\x99\xa6',            r'$\diamondsuit$'),
        (MODE_REPLACE,  '\xe2\x88\x8d',            r'$\ni$'),
        (MODE_REPLACE,  '\xe2\x86\x92',            r'$\rightarrow$'),
        (MODE_REPLACE,  '\xce\x9e',                r'$\Xi$'),
        (MODE_REPLACE,  '\xce\xb9',                r'$\iota$'),
        (MODE_REPLACE,  '\xce\xbd',                r'$\nu$'),
        (MODE_REPLACE,  '\xc3\xb7',                r'$\div$'),
        (MODE_REPLACE,  '\xc2\xa7',                r'\S'),
        (MODE_REPLACE,  '\xcf\x91',                r'$\vartheta$'),
        (MODE_REPLACE,  '\xcf\x86',                r'$\phi$'),
        (MODE_REPLACE,  '\xe2\x8a\xa2',            r'$\vdash$'),
        (MODE_REPLACE,  '\xe2\x88\x89',            r'$\notin$'),
        (MODE_REPLACE,  '\xe2\x8a\x96',            r'$\ominus$'),
        (MODE_REPLACE,  '\xe2\x84\x8f',            r'$\hbar$'),
        (MODE_REPLACE,  '\xe2\x88\xa0',            r'$\angle$'),
        (MODE_REPLACE,  '\xe2\x89\xbb',            r'$\succ$'),
        (MODE_REPLACE,  '\xe2\x88\x9e',            r'$\infty$'),
        (MODE_REPLACE,  '\xce\xbe',                r'$\xi$'),
        (MODE_REPLACE,  '\xe2\x99\xa5',            r'$\heartsuit$'),
        (MODE_REPLACE,  '\xe2\x89\x89',            r'$\not\approx$'),
        (MODE_REPLACE,  '\xc2\xb6',                r'\P'),
        (MODE_REPLACE,  '\xe2\x87\x90',            r'$\Leftarrow$'),
        (MODE_REPLACE,  '\xe2\x88\xa8',            r'$\vee$'),
        (MODE_REPLACE,  '\xce\xb3',                r'$\gamma$'),
        (MODE_REPLACE,  '\xe2\x86\x95',            r'$\updownarrow$'),
        (MODE_REPLACE,  '\xe2\x89\x8d',            r'$\asymp$'),
        (MODE_REPLACE,  '\xe2\x88\xab',            r'$\int$'),
        (MODE_REPLACE,  '\xe2\x88\xbc',            r'$\sim$'),
        (MODE_REPLACE,  '\xe2\x99\xa7',            r'$\clubsuit$'),
        (MODE_REPLACE,  '\xe2\x85\x9e',            r'$\frac78$'),
        (MODE_REPLACE,  '\xe2\x86\xa9',            r'$\hookleftarrow$'),
        (MODE_REPLACE,  '\xe2\x86\x93',            r'$\downarrow$'),
        (MODE_REPLACE,  '\xe2\x89\xab',            r'$\gg$'),
        (MODE_REPLACE,  '\xe2\x8b\x81',            r'$\bigvee$'),
        (MODE_REPLACE,  '\xe2\x89\xad',            r'$\not\asymp$'),
        (MODE_REPLACE,  '\xe2\x87\x95',            r'$\Updownarrow$'),
        (MODE_REPLACE,  '\xce\xb4',                r'$\delta$'),
        (MODE_REPLACE,  '\xe2\x86\xbc',            r'$\leftharpoonup$'),
        (MODE_REPLACE,  '\xe2\x89\x83',            r'$\simeq$'),
        (MODE_REPLACE,  '\xe2\x88\xa5',            r'$\parallel$'),
        (MODE_REPLACE,  '\xe2\x85\x9d',            r'$\frac58$'),
        (MODE_REPLACE,  '\xe2\x85\x9a',            r'$\frac56$'),
        (MODE_REPLACE,  '\xe2\x89\x88',            r'$\approx$'),
        (MODE_REPLACE,  '\xe2\x87\x81',            r'$\rightharpoondown$'),
        (MODE_REPLACE,  '\xe2\x88\x82',            r'$\partial$'),
        (MODE_REPLACE,  '\xe2\x8a\x86',            r'$\subseteq$'),
        (MODE_REPLACE,  '\xe2\x87\x8c',            r'$\rightleftharpoons$'),
        (MODE_REPLACE,  '\xe2\x88\xa3',            r'$\mid$'),
        (MODE_REPLACE,  '\xe2\x85\x9f',            r'$\frac14$'),
        (MODE_REPLACE,  '\xe2\x86\xbd',            r'$\leftharpoondown$'),
        (MODE_REPLACE,  '\xe2\x87\x93',            r'$\Downarrow$'),
        (MODE_REPLACE,  '\xe2\x8b\x88',            r'$\bowtie$'),
        (MODE_REPLACE,  '\xe2\x99\xae',            r'$\natural$'),
        (MODE_REPLACE,  '\xe2\x8b\xae',            r'$\vdots$'),
        (MODE_REPLACE,  '\xe2\x88\x91',            r'$\sum$'),
        (MODE_REPLACE,  '\xe2\x8a\x85',            r'$\not\supset$'),
        (MODE_REPLACE,  '\xe2\x8a\x82',            r'$\subset$'),

        # Various abbreviation replacements...
        # TODO: make regexp patterns or subconverter
        (MODE_REGEXP,   r'\\S[ ]{0,1}(\d)',         r'\S\,\1'),
        (MODE_REPLACE,  'e.g.',                    'e.\,g.'),
        (MODE_REPLACE,  'e. g.',                   'e.\,g.'),
        (MODE_REPLACE,  'i.d.r',                   'i.\,d.\,r.'),
        (MODE_REPLACE,  'i. d. r.',                'i.\,d.\,r.'),
        (MODE_REPLACE,  'z.B.',                    'z.\,B.'),
        (MODE_REPLACE,  'z. B.',                   'z.\,B.'),
        (MODE_REPLACE,  'lic.iur.',                'lic.\,iur.'),
        (MODE_REPLACE,  'Dr.iur.',                 'Dr.\,iur.'),

        (MODE_REGEXP,   r'S\. (\d+)',             r'S.~\1'),
        (MODE_REGEXP,   r'Abs\. (\d+)',           r'Abs.~\1'),
        (MODE_REGEXP,   r'(\d+) m2',              r'\1~m2'),
        (MODE_REGEXP,   r'Art\. (\d+)',           r'Art.~\1'),
        (MODE_REGEXP,   r'SR (\d+)',              r'SR~\1'),

        # Use dashes and non breaking space in swiss currencies
        (MODE_REGEXP, r'Fr\.[ ~](\d{1,})\.(%s|-{1,2})' % DASH,
                      r'Fr.~\1.--'),


        # date replacements: we use nonbreakable spaces
        # TODO: make regexp patterns or subconverter
        ] + [
        (MODE_REGEXP,    r'(\d)\.\W{0,1}(' + month + ')',
         r'\1.~\2')
        for month in ['Jan', 'Feb', 'Mar', 'Mär', 'Apr', 'Mai', 'May', 'Jun',
                      'Jul', 'Aug', 'Sep', 'Okt', 'Oct', 'Nov', 'Dez', 'Dec']
        ] + [

        # NOT SUPPORTED!!
        (MODE_REPLACE,  '&yen;',                   ''),
        (MODE_REPLACE,  '&brvbar;',                ''),
        (MODE_REPLACE,  '‛',                       ''),
        (MODE_REPLACE,  '&diams;',                 ''),
        (MODE_REPLACE,  '►',                       ''),

        # misc
        (MODE_REGEXP,   r'([a-zA-Z])-,',           r'\1"~,'),
        (MODE_REGEXP,   r'([a-zA-Z])-',            r'\1"='),
        # - at the end of a word should not disable hyphenation

        #    (MODE_REGEXP,   r'([a-zA-Z].)*',           r'U\1U'),
        # abbreviation should contain short whitespace (\,)

        # cleanup
        # strip html tags
        (MODE_REGEXP,   r'<([^>]*?)>',             r''),
        # remove all line endings (\\) before \sub*
        (MODE_REGEXP,   r'\\\\\s*(\n\\sub)',       r'\g<1>'),
        # remove all line endings (\\) which finish a empty line
        (MODE_REGEXP,   r'(\n|^)[ ]{0,}\\\\',      r'\g<1>'),
        # remove all line endings (\\) after a } (latex command)
        (MODE_REGEXP,   r'}[\t ]{0,}\\\\',         r'}'),

        # special characters
        (MODE_REPLACE,  '&lt;',                    '<'),
        (MODE_REPLACE,  '&gt;',                    '>'),
        (MODE_REPLACE,  '&',                       '\\&'),
        interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_BOTTOM,
        (MODE_REPLACE,  interfaces.HTML2LATEX_PREVENT_CHARACTER,         ''),
        (MODE_REPLACE,  BACKSLASH_MARKER,
         r'\textbackslash '),
        ])



from htmlentitydefs import name2codepoint as n2cp
from htmlentitydefs import codepoint2name as cp2n
import re
from zope.interface import directlyProvidedBy, directlyProvides


def provide_request_layer(request, layer):
    """ Add a layer interface on the request
    """
    layer = type(layer) == list and layer or [layer]
    ifaces = layer + list(directlyProvidedBy(request))

    # Since we allow multiple markers here, we can't use
    # zope.publisher.browser.applySkin() since this filters out
    # IBrowserSkinType interfaces, nor can we use alsoProvides(), since
    # this appends the interface (in which case we end up *after* the
    # default Plone/CMF skin)
    directlyProvides(request, *ifaces)


def baseclasses(cls, bases=None):
    """Returns a flat list of all baseclasses according to the method
    resolution order. Includes `cls`.
    """
    if bases is None:
        bases = []

    bases.append(cls)

    for base in cls.__bases__:
        if base not in bases:
            baseclasses(base, bases)

    return bases


def decode_htmlentities(string):
    """
    Decodes html entities and xml entities.
    """
    entity_re = re.compile("&(#?)(\d{1,5}|\w{1,8});")

    def substitute_entity(match):
        ent = match.group(2)
        if match.group(1) == "#":
            return unichr(int(ent))

        else:
            cp = n2cp.get(ent)

            if cp:
                return unichr(cp)

            else:
                return match.group()

    return entity_re.subn(substitute_entity, string)[0]


def encode_htmlentities(string, encoding='utf-8'):
    """
    Encodes the string with html entities.
    """

    if isinstance(string, unicode):
        was_unicode = True
    else:
        was_unicode = False
        string = string.decode(encoding)

    string = string.replace('&', '&amp;')

    for codepoint, name in cp2n.items():
        if name != 'amp':
            string = string.replace(unichr(codepoint), '&%s;' % name)

    if not was_unicode:
        string = string.encode(encoding)
    return string


def html2xmlentities(string):
    """
    Converts htmlentities to xmlentities
    """

    xpr = re.compile('&(\w{1,8});')

    def substitute_entity(match):
        ent = match.group(1)
        if ent in n2cp.keys():
            return '&#%i;' % n2cp[ent]

        else:
            return match.group(0)

    return xpr.subn(substitute_entity, string)[0]


def xml2htmlentities(string):
    """
    Converts xmlentities to htmlentities
    """
    xpr = re.compile('&#(\d{1,5});')

    def substitute_entity(match):
        ent = int(match.group(1))
        if ent in cp2n.keys():
            return '&%s;' % cp2n[ent]

        else:
            return match.group(0)

    return xpr.subn(substitute_entity, string)[0]
